// Part of DTSource. Copyright 2004-2012. David Adalsteinsson.  BSD License
// see http://www.visualdatatools.com/DTSource/license.html for more information.

#include "DTDoubleArray.h"
#include "DTError.h"
#include "DTIntArray.h"
#include "DTList.h"

#include "DTArrayTemplates.h"

#include <math.h>
#include <string.h>
#include <algorithm>
#include <limits>

#if !defined(INFINITY)
#if defined(WIN32)
#define INFINITY std::numeric_limits<float>::infinity();
#else
#define INFINITY 1e50f
#endif
#endif

DTDoubleArrayStorage::DTDoubleArrayStorage(long int mv,long int nv,long int ov)
{
    // Check if it's called correctly.
    m = mv>0 ? mv : 0;
    n = nv>0 ? nv : 0;
    o = ov>0 ? ov : 0;
    length = m*n*o;
    if (length==0) m = n = o = 0;
    referenceCount = 1;
    mn = m*n;
    mutableReferences = 0;
    
    Data = length==0 ? NULL : new double[length];
}

DTDoubleArrayStorage::~DTDoubleArrayStorage()
{
    delete [] Data;
}

DTDoubleArray::~DTDoubleArray()
{
    if (Storage) {
        Storage->accessLock.Lock();
        int refCnt = (--Storage->referenceCount);
        Storage->accessLock.Unlock();
        if (refCnt==0) delete Storage;
    }
}

DTDoubleArray::DTDoubleArray(const DTDoubleArray &A) 
{
    Storage = A.Storage;
    Storage->accessLock.Lock();
    Storage->referenceCount++;
    Storage->accessLock.Unlock();
}

DTDoubleArray &DTDoubleArray::operator=(const DTDoubleArray &A)
{
    if (Storage!=A.Storage) {
        Storage->accessLock.Lock();
        A.Storage->accessLock.Lock();
        Storage->referenceCount--;
        int refCnt = Storage->referenceCount;
        Storage->accessLock.Unlock();
        if (refCnt==0) delete Storage;
        Storage = A.Storage;
        Storage->referenceCount++;
        Storage->accessLock.Unlock();
    }
    
     return *this;
}

int DTDoubleArray::ReferenceCount() const
{
    Storage->accessLock.Lock(); 
    int refCnt = Storage->referenceCount;
    Storage->accessLock.Unlock();
    return refCnt;
}

int DTDoubleArray::MutableReferences() const
{
    Storage->accessLock.Lock(); 
    int toReturn = Storage->mutableReferences;
    Storage->accessLock.Unlock();
    return toReturn;
}

long int DTDoubleArray::m() const
{
    return Storage->m;
}

long int DTDoubleArray::n() const
{
    return Storage->n;
}

long int DTDoubleArray::o() const
{
    return Storage->o;
}

long int DTDoubleArray::Length() const
{
    return Storage->length;
}

bool DTDoubleArray::IsEmpty() const
{
    return (Storage->length==0);
}

bool DTDoubleArray::NotEmpty() const
{
    return (Storage->length!=0);
}

double DTDoubleArray::e(int i) const
{
    if (i<0 || i>=Storage->length) {
        cerr << "Out of bounds" << endl;
        return invalidEntry;
    }
    else
        return Storage->Data[i];
}

double DTDoubleArray::e(int i,int j) const
{
    if (i<0 || i>=Storage->m || j<0 || j>=Storage->n) {
        cerr << "Out of bounds" << endl;
        return invalidEntry;
    }
    else
        return Storage->Data[i+j*Storage->m];
}

double DTDoubleArray::e(int i,int j,int k) const
{
    if (i<0 || i>=Storage->m || j<0 || j>=Storage->n || k<0 || k>=Storage->o) {
        cerr << "Out of bounds" << endl;
        return invalidEntry;
    }
    else
        return Storage->Data[i+j*Storage->m+k*Storage->mn];
}

void DTDoubleArray::pinfo(void) const
{
    if (o()==0)
        cerr << "Empty" << endl;
    else if (o()==1) {
        if (n()==1) {
            if (m()==1) {
                cerr << "double array with one entry" << endl;
            }
            else {
                cerr << "double vector with " << m() << " entries" << endl;
            }
        }
        else
            cerr << m() << " x " << n() << " double array" << endl;
    }
    else
        cerr << m() << " x " << n() << " x " << o() << " double array" << endl;
    
    cerr.flush();
}

void DTDoubleArray::pi(int i) const
{
    if (i<0 || i>=m()) {
        cerr << "Out of bounds." << endl;
    }
    else {
        long int howMany = n();
        long int j;
        for (j=0;j<howMany-1;j++) cerr << operator()(i,j) << ", ";
        if (howMany>0) cerr << operator()(i,howMany-1);
        cerr << endl;
    }
}

void DTDoubleArray::pj(int j) const
{
    if (j<0 || j>=n()) {
        cerr << "Out of bounds." << endl;
    }
    else {
        long int howMany = m();
        long int i;
        for (i=0;i<howMany-1;i++) cerr << operator()(i,j) << ", ";
        if (howMany>0) cerr << operator()(howMany-1,j);
        cerr << endl;
    }
}

void DTDoubleArray::pall(void) const
{
    long int mv = m();
    long int nv = n();
    long int i,j;
    if (mv==0) {
        cerr << "Empty" << endl;
    }
    else {
        for (j=0;j<nv;j++) {
            for (i=0;i<mv-1;i++) cerr << operator()(i,j) << ", ";
            cerr << operator()(mv-1,j);
            cerr << endl;
        }
    }
}

void DTDoubleArray::prange(int startIndex,int endIndex) const
{
    // Offsets [s,e], both included.
    long int i;
    if (startIndex<0) {
        cerr << "start out of bounds" << endl;
        return;
    }
    if (endIndex>=Length()) {
        cerr << "end out of bounds" << endl;
        return;
    }
    for (i=startIndex;i<endIndex;i++) {
        cerr << operator()(i) << ", ";
    }
    if (startIndex<=endIndex) {
        cerr << operator()(i) << endl;
    }
}

void DTDoubleArray::pcrange(int startIndex,int endIndex) const
{
    // Print (:,[s:e]), otherwise like pall().
    long int mv = m();
    long int nv = n();
    long int i,j;
    if (startIndex<0) {
        cerr << "start out of bounds" << endl;
        return;
    }
    if (endIndex>=nv) {
        cerr << "end out of bounds" << endl;
        return;
    }
    
    for (j=startIndex;j<=endIndex;j++) {
        for (i=0;i<mv-1;i++) cerr << operator()(i,j) << ", ";
        cerr << operator()(mv-1,j);
        cerr << endl;
    }
}

void DTDoubleArray::psigns(void) const
{
    long int mv = m();
    long int nv = n();
    long int i,j;
    if (mv==0) {
        cerr << "Empty" << endl;
    }
    else if (o()>1) {
        cerr << "only works for 2D arrays" << endl;
    }
    else {
		double val;
        for (j=0;j<nv;j++) {
            for (i=0;i<mv;i++) {
				val = operator()(i,j);
				if (isfinite(val)==0) {
					if (isnan(val))
						cerr << "*";
					else if (val<0)
						cerr << "P";
					else if (val>0)
						cerr << "M";
				}
				else if (val<0)
					cerr << "-";
				else if (val==0)
					cerr << "0";
				else
					cerr << "+";
			}
            cerr << endl;
        }
    }
}

long int DTDoubleArray::Find(double v) const
{
    const double *D = Pointer();
    long int len = Length();
    long int i;
    for (i=0;i<len;i++) {
        if (D[i]==v) break;
    }

    return (i<len ? i : -1);
}

DTMutableDoubleArray DTDoubleArray::Copy() const
{
    DTMutableDoubleArray CopyInto(m(),n(),o());
    // Check that the allocation worked.
    if (CopyInto.Length()!=Length()) return CopyInto; // Failed.  Already printed an error message.
    memcpy(CopyInto.Pointer(),Pointer(),Length()*sizeof(double));
    return CopyInto;
}

void DTDoubleArray::PrintErrorMessage(long int i) const
{
    DTErrorOutOfRange("DTDoubleArray",i,Storage->length);
}

void DTDoubleArray::PrintErrorMessage(long int i,long int j) const
{
    DTErrorOutOfRange("DTDoubleArray",i,j,Storage->m,Storage->n);
}

void DTDoubleArray::PrintErrorMessage(long int i,long int j,long int k) const
{
    DTErrorOutOfRange("DTDoubleArray",i,j,k,Storage->m,Storage->n,Storage->o);
}

DTMutableDoubleArray::DTMutableDoubleArray(const DTMutableDoubleArray &A)
: DTDoubleArray(A)
{
    Storage->accessLock.Lock();
    Storage->mutableReferences++;
    Storage->accessLock.Unlock();
}

DTMutableDoubleArray::~DTMutableDoubleArray()
{
    Storage->accessLock.Lock();
    int refCnt = (--Storage->referenceCount);
    Storage->mutableReferences--;
    Storage->accessLock.Unlock();
    if (refCnt==0) delete Storage;
    Storage = NULL;
}

DTMutableDoubleArray &DTMutableDoubleArray::operator=(const DTMutableDoubleArray &A)
{
    if (Storage!=A.Storage) {
        Storage->accessLock.Lock();
        A.Storage->accessLock.Lock();
        Storage->referenceCount--;
        int refCnt = Storage->referenceCount;
        Storage->mutableReferences--;
        Storage->accessLock.Unlock();
        if (refCnt==0) delete Storage;
        Storage = A.Storage;
        Storage->referenceCount++;
        Storage->mutableReferences++;
        Storage->accessLock.Unlock();
    }

    return *this;
}

DTMutableDoubleArray &DTMutableDoubleArray::operator=(double a)
{
    const long int howManyNumbers = Length();
    if (a==0.0) {
        memset(Pointer(),0,sizeof(double)*howManyNumbers);
    }
    else {
//#ifndef _ANSI_SOURCE
//        memset_pattern8(Pointer(),&a,sizeof(double)*howManyNumbers);
//#else
        long int i;
        double *Data = Pointer();
        for (i=0;i<howManyNumbers;i++)
            Data[i] = a;
//#endif
    }
    
    return *this;
}

void DTMutableDoubleArray::operator*=(double v)
{
    DTTimesEqualsScalar<DTMutableDoubleArray,double>(*this,v);
}

void DTMutableDoubleArray::operator/=(double v)
{
    DTDivideEqualsScalar<DTMutableDoubleArray,double>(*this,v);
}

void DTMutableDoubleArray::operator+=(double v)
{
    DTPlusEqualsScalar<DTMutableDoubleArray,double>(*this,v);
}

void DTMutableDoubleArray::operator-=(double v)
{
    DTMinusEqualsScalar<DTMutableDoubleArray,double>(*this,v);
}

void DTMutableDoubleArray::operator+=(const DTDoubleArray &B)
{
    DTPlusEqualsArray<DTDoubleArray,DTMutableDoubleArray,double>(*this,B);
}

void DTMutableDoubleArray::operator-=(const DTDoubleArray &B)
{
    DTMinusEqualsArray<DTDoubleArray,DTMutableDoubleArray,double>(*this,B);
}

void DTMutableDoubleArray::operator*=(const DTDoubleArray &B)
{
    DTTimesEqualsArray<DTDoubleArray,DTMutableDoubleArray,double>(*this,B);
}

void DTMutableDoubleArray::operator/=(const DTDoubleArray &B)
{
    DTDivideEqualsArray<DTDoubleArray,DTMutableDoubleArray,double>(*this,B);
}

bool operator==(const DTDoubleArray &A,const DTDoubleArray &B)
{
    return DTOperatorArrayEqualsArray<DTDoubleArray,double>(A,B);
}

bool operator!=(const DTDoubleArray &A,const DTDoubleArray &B)
{
    return !(A==B);
}

DTMutableDoubleArray TruncateSize(const DTDoubleArray &A,long int length)
{
    return DTTruncateArraySize<DTDoubleArray,DTMutableDoubleArray,double>(A,length);
}

DTMutableDoubleArray IncreaseSize(const DTDoubleArray &A,long int addLength)
{
    return DTIncreaseArraySize<DTDoubleArray,DTMutableDoubleArray,double>(A,addLength);
}

DTMutableDoubleArray Transpose(const DTDoubleArray &A)
{
    return DTTransposeArray<DTDoubleArray,DTMutableDoubleArray,double>(A);
}

DTMutableDoubleArray Reshape(const DTDoubleArray &A,long int m,long int n,long int o)
{
    if (m<0 || n<0 || o<0) {
        DTErrorMessage("Reshape(DTDoubleArray,...)","One of the new dimensions is negative.");
        return DTMutableDoubleArray();
    }
    if (m*n*o!=A.Length()) {
        DTErrorMessage("Reshape(DTDoubleArray,...)","Size before and after need to be the same.");
        return DTMutableDoubleArray();
    }
    
    DTMutableDoubleArray toReturn(m,n,o);
    if (toReturn.Length()) {
        memcpy(toReturn.Pointer(),A.Pointer(),A.Length()*sizeof(double));
    }
    
    return toReturn;
}

DTMutableDoubleArray Sort(const DTDoubleArray &A)
{
    DTMutableDoubleArray toReturn = Reshape(A,A.Length());
    sort(toReturn.Pointer(),toReturn.Pointer()+toReturn.Length());
    return toReturn;
}

struct DTArraySortingPair {
    int operator<(const DTArraySortingPair &a) const {return (v<a.v);}
    double v;
	int i;
};

DTMutableIntArray SortedOrder(const DTDoubleArray &A)
{
	// A(returned(i)) = i'th entry in a sorted version of A.  So A(SortedOrder(A)) = Sort(A)
	int i,howMany = A.Length();
	DTMutableList<DTArraySortingPair> list(howMany);
	for (i=0;i<howMany;i++) {
		list(i).v = A(i);
		list(i).i = i;
	}
	sort(list.Pointer(),list.Pointer()+howMany);
	DTMutableIntArray toReturn(howMany);
	for (i=0;i<howMany;i++) {
		toReturn(i) = list(i).i;
	}
	return toReturn;
}

DTMutableDoubleArray FlipJ(const DTDoubleArray &A)
{
    return DTArrayFlipJ<DTDoubleArray,DTMutableDoubleArray,double>(A);
}

void Swap(DTMutableDoubleArray &A,DTMutableDoubleArray &B)
{
	DTMutableDoubleArray C = A;
	A = B;
	B = C;
}

void Swap(DTDoubleArray &A,DTDoubleArray &B)
{
	DTDoubleArray C = A;
	A = B;
	B = C;
}

void CopyValues(DTMutableDoubleArray &into,const DTDoubleArray &from)
{
	if (into.m()!=from.m() || into.n()!=from.n() || into.o()!=from.o()) {
		DTErrorMessage("CopyValues(MutableDoubleArray,DoubleArray)","Incompatible sizes");
	}
	else if (into.NotEmpty()) {
		memcpy(into.Pointer(),from.Pointer(),into.Length()*sizeof(double));
	}
}

void CopyIntoColumn(DTMutableDoubleArray &into,const DTDoubleArray &list,int j)
{
    if (into.m()!=list.Length()) {
		DTErrorMessage("CopyIntoColumn(into,list,j)","Need into.m()==list.Length()");
    }
    else if (into.o()!=1) {
		DTErrorMessage("CopyIntoColumn(into,list,j)","into is a 3D array (into.o()>1)");
    }
    else if (j<0 || j>into.n()) {
        DTErrorMessage("CopyIntoColumn(into,list,j)","j out of bounds");
    }
    else {
        memcpy(into.Pointer()+j*into.m(),list.Pointer(),into.m()*sizeof(double));
    }
}

DTMutableDoubleArray SubArray(const DTDoubleArray &A,int imin,int icount,int jmin,int jcount,int kmin,int kcount)
{
	if (imin<0 || imin+icount>A.m() || jmin<0 || jmin+jcount>A.n() || kmin<0 || kmin+kcount>A.o()) {
		DTErrorMessage("SubArray(A,....)","Index out of range");
	}
	if (icount<=0 || jcount<=0 || kcount<=0) return DTMutableDoubleArray();
	DTMutableDoubleArray toReturn(icount,jcount,kcount);
	int j,k;
	double *into = toReturn.Pointer();
	const double *from = A.Pointer();
	const int m = A.m();
	const int n = A.n();
	for (k=0;k<kcount;k++) {
		for (j=0;j<jcount;j++) {
			if (icount==1)
				toReturn(0,j,k) = A(imin,jmin+j,kmin+k);
			else
				memcpy(into+j*icount+k*icount*jcount,from + imin + (jmin+j)*m+(kmin+k)*m*n, icount*sizeof(double));
		}
	}
	
	return toReturn;
}

double Minimum(const DTDoubleArray &A)
{
    long int len = A.Length();
    double minV = INFINITY;
    
    double v;
    long int i;
    
    const double *D = A.Pointer();
    
    for (i=0;i<len;i++) {
        v = D[i];
        minV = (v < minV ? v : minV);
    }

    return minV;
}

double Maximum(const DTDoubleArray &A)
{
    long int len = A.Length();
    double maxV = -INFINITY;
    
    double v;
    long int i;
    
    const double *D = A.Pointer();
    
    for (i=0;i<len;i++) {
        v = D[i];
        maxV = (maxV < v ? v : maxV); // Will ignore NAN values
    }
    
    return maxV;
}

DTMutableDoubleArray CombineColumns(const DTDoubleArray &First,const DTDoubleArray &Second)
{
    if (First.m()!=Second.m()) {
        DTErrorMessage("CombineColumns(A,B)","A and B have to have the same number of rows.");
        return DTMutableDoubleArray();
    }
    if (First.IsEmpty())
        return DTMutableDoubleArray();
    if (First.o()!=1 || Second.o()!=1) {
        DTErrorMessage("CombineColumns(A,B)","A and B have to be two dimensional.");
        return DTMutableDoubleArray();
    }
    
    DTMutableDoubleArray toReturn(First.m(),First.n()+Second.n());
    memcpy(toReturn.Pointer(),First.Pointer(),First.Length()*sizeof(double));
    memcpy(toReturn.Pointer()+First.Length(),Second.Pointer(),Second.Length()*sizeof(double));
    
    return toReturn;
}

