// Part of DTSource. Copyright 2008-2012. David Adalsteinsson.  BSD License
// see http://www.visualdatatools.com/DTSource/license.html for more information.

#include "DTDoubleComplexArray.h"


#include "DTDoubleComplexArray.h"
#include "DTDoubleArray.h"
#include "DTError.h"

#include "DTArrayTemplates.h"

#include <math.h>
#include <string.h>
#include <algorithm>
#include <limits>

#if !defined(INFINITY)
#if defined(WIN32)
#define INFINITY std::numeric_limits<float>::infinity();
#else
#define INFINITY 1e50f
#endif
#endif

DTDoubleComplexArrayStorage::DTDoubleComplexArrayStorage(long int mv,long int nv,long int ov)
{
    // Check if it's called correctly.
    m = mv>0 ? mv : 0;
    n = nv>0 ? nv : 0;
    o = ov>0 ? ov : 0;
    length = m*n*o;
    if (length==0) m = n = o = 0;
    referenceCount = 1;
    mn = m*n;
    mutableReferences = 0;
    
    Data = length==0 ? NULL : new DTDoubleComplex[length];
}

DTDoubleComplexArrayStorage::~DTDoubleComplexArrayStorage()
{
    delete Data;
}

DTDoubleComplexArray::~DTDoubleComplexArray()
{
    if (Storage) {
        accessLock.Lock();
        Storage->accessLock.Lock();
        int refCnt = (--Storage->referenceCount);
        Storage->accessLock.Unlock();
        if (refCnt==0) delete Storage;
        accessLock.Unlock();
    }
}

DTDoubleComplexArray::DTDoubleComplexArray(const DTDoubleComplexArray &A) 
{
    accessLock.Lock();
    Storage = A.Storage;
    Storage->accessLock.Lock();
    Storage->referenceCount++;
    Storage->accessLock.Unlock();
    accessLock.Unlock();
}

DTDoubleComplexArray &DTDoubleComplexArray::operator=(const DTDoubleComplexArray &A)
{
    if (accessLock==A.accessLock) {
        // A=A, safe but pointless.
        return *this;
    }
    
    // Allow A = A
    accessLock.Lock();
    A.accessLock.Lock();
    
    if (Storage!=A.Storage) {
        Storage->accessLock.Lock();
        A.Storage->accessLock.Lock();
        Storage->referenceCount--;
        int refCnt = Storage->referenceCount;
        Storage->accessLock.Unlock();
        if (refCnt==0) delete Storage;
        Storage = A.Storage;
        Storage->referenceCount++;
        Storage->accessLock.Unlock();
    }
    
    A.accessLock.Unlock();
    accessLock.Unlock();
    
    return *this;
}

int DTDoubleComplexArray::ReferenceCount() const
{
    accessLock.Lock();
    Storage->accessLock.Lock(); 
    int refCnt = Storage->referenceCount;
    Storage->accessLock.Unlock();
    accessLock.Unlock();
    return refCnt;
}

int DTDoubleComplexArray::MutableReferences() const
{
    accessLock.Lock();
    Storage->accessLock.Lock(); 
    int toReturn = Storage->mutableReferences;
    Storage->accessLock.Unlock();
    accessLock.Unlock();
    return toReturn;
}

long int DTDoubleComplexArray::m() const
{
    return Storage->m;
}

long int DTDoubleComplexArray::n() const
{
    return Storage->n;
}

long int DTDoubleComplexArray::o() const
{
    return Storage->o;
}

long int DTDoubleComplexArray::Length() const
{
    return Storage->length;
}

bool DTDoubleComplexArray::IsEmpty() const
{
    return (Storage->length==0);
}

bool DTDoubleComplexArray::NotEmpty() const
{
    return (Storage->length!=0);
}

void DTDoubleComplexArray::pinfo(void) const
{
    if (o()==0)
        cerr << "Empty" << endl;
    else if (o()==1) {
        if (n()==1) {
            if (m()==1) {
                cerr << "complex array with one entry" << endl;
            }
            else {
                cerr << "complex vector with " << m() << " entries" << endl;
            }
        }
        else
            cerr << m() << " x " << n() << " complex array" << endl;
    }
    else
        cerr << m() << " x " << n() << " x " << o() << " complex array" << endl;
    
    cerr << flush;
}

long int DTDoubleComplexArray::Find(DTDoubleComplex v) const
{
    const DTDoubleComplex *D = Pointer();
    long int len = Length();
    long int i;
    for (i=0;i<len;i++) {
        if (D[i]==v) break;
    }

    return (i<len ? i : -1);
}

DTMutableDoubleComplexArray DTDoubleComplexArray::Copy() const
{
    DTMutableDoubleComplexArray CopyInto(m(),n(),o());
    // Check that the allocation worked.
    if (CopyInto.Length()!=Length()) return CopyInto; // Failed.  Already printed an error message.
    memcpy(CopyInto.Pointer(),Pointer(),Length()*sizeof(DTDoubleComplex));
    return CopyInto;
}

void DTDoubleComplexArray::PrintErrorMessage(long int i) const
{
    DTErrorOutOfRange("DTDoubleComplexArray",i,Storage->length);
}

void DTDoubleComplexArray::PrintErrorMessage(long int i,long int j) const
{
    DTErrorOutOfRange("DTDoubleComplexArray",i,j,Storage->m,Storage->n);
}

void DTDoubleComplexArray::PrintErrorMessage(long int i,long int j,long int k) const
{
    DTErrorOutOfRange("DTDoubleComplexArray",i,j,k,Storage->m,Storage->n,Storage->o);
}

DTMutableDoubleComplexArray::DTMutableDoubleComplexArray(const DTMutableDoubleComplexArray &A)
{
    accessLock.Lock();
    Storage = A.Storage;
    Storage->accessLock.Lock();
    Storage->referenceCount++;
    Storage->mutableReferences++;
    Storage->accessLock.Unlock();
    accessLock.Unlock();
}

DTMutableDoubleComplexArray::~DTMutableDoubleComplexArray()
{
    accessLock.Lock();
    Storage->accessLock.Lock();
    int refCnt = (--Storage->referenceCount);
    Storage->mutableReferences--;
    Storage->accessLock.Unlock();
    if (refCnt==0) delete Storage;
    Storage = NULL;
    accessLock.Unlock();
}

DTMutableDoubleComplexArray &DTMutableDoubleComplexArray::operator=(const DTMutableDoubleComplexArray &A)
{
    if (accessLock==A.accessLock) {
        // A=A, safe but pointless.
        return *this;
    }
    
    // Allow A = A
    accessLock.Lock();
    A.accessLock.Lock();
    
    if (Storage!=A.Storage) {
        Storage->accessLock.Lock();
        A.Storage->accessLock.Lock();
        Storage->referenceCount--;
        int refCnt = Storage->referenceCount;
        Storage->mutableReferences--;
        Storage->accessLock.Unlock();
        if (refCnt==0) delete Storage;
        Storage = A.Storage;
        Storage->referenceCount++;
        Storage->mutableReferences++;
        Storage->accessLock.Unlock();
    }
    
    A.accessLock.Unlock();
    accessLock.Unlock();
    
    return *this;
}

DTMutableDoubleComplexArray &DTMutableDoubleComplexArray::operator=(DTDoubleComplex a)
{
    const long int howManyNumbers = Length();
    if (a==0.0) {
        memset(Pointer(),0,sizeof(DTDoubleComplex)*howManyNumbers);
    }
    else {
        long int i;
        DTDoubleComplex *Data = Pointer();
        for (i=0;i<howManyNumbers;i++)
            Data[i] = a;
    }
    
    return *this;
}

bool operator==(const DTDoubleComplexArray &A,const DTDoubleComplexArray &B)
{
    return DTOperatorArrayEqualsArray<DTDoubleComplexArray,DTDoubleComplex>(A,B);
}

bool operator!=(const DTDoubleComplexArray &A,const DTDoubleComplexArray &B)
{
    return !(A==B);
}

DTMutableDoubleComplexArray TruncateSize(const DTDoubleComplexArray &A,long int length)
{
    return DTTruncateArraySize<DTDoubleComplexArray,DTMutableDoubleComplexArray,DTDoubleComplex>(A,length);
}

DTMutableDoubleComplexArray IncreaseSize(const DTDoubleComplexArray &A,long int addLength)
{
    return DTIncreaseArraySize<DTDoubleComplexArray,DTMutableDoubleComplexArray,DTDoubleComplex>(A,addLength);
}

DTMutableDoubleComplexArray Transpose(const DTDoubleComplexArray &A)
{
    return DTTransposeArray<DTDoubleComplexArray,DTMutableDoubleComplexArray,DTDoubleComplex>(A);
}

DTMutableDoubleComplexArray Reshape(const DTDoubleComplexArray &A,long int m,long int n,long int o)
{
    if (m<0 || n<0 || o<0) {
        DTErrorMessage("Reshape(DTDoubleComplexArray,...)","One of the new dimensions is negative.");
        return DTMutableDoubleComplexArray();
    }
    if (m*n*o!=A.Length()) {
        DTErrorMessage("Reshape(DTDoubleComplexArray,...)","Size before and after need to be the same.");
        return DTMutableDoubleComplexArray();
    }
    
    DTMutableDoubleComplexArray toReturn(m,n,o);
    if (toReturn.Length()) {
        memcpy(toReturn.Pointer(),A.Pointer(),A.Length()*sizeof(DTDoubleComplex));
    }
    
    return toReturn;
}

DTMutableDoubleComplexArray FlipJ(const DTDoubleComplexArray &A)
{
    return DTArrayFlipJ<DTDoubleComplexArray,DTMutableDoubleComplexArray,DTDoubleComplex>(A);
}

DTMutableDoubleComplexArray CombineColumns(const DTDoubleComplexArray &First,const DTDoubleComplexArray &Second)
{
    if (First.m()!=Second.m()) {
        DTErrorMessage("CombineColumns(A,B)","A and B have to have the same number of rows.");
        return DTMutableDoubleComplexArray();
    }
    if (First.IsEmpty())
        return DTMutableDoubleComplexArray();
    if (First.o()!=1 || Second.o()!=1) {
        DTErrorMessage("CombineColumns(A,B)","A and B have to be two dimensional.");
        return DTMutableDoubleComplexArray();
    }
    
    DTMutableDoubleComplexArray toReturn(First.m(),First.n()+Second.n());
    memcpy(toReturn.Pointer(),First.Pointer(),First.Length()*sizeof(DTDoubleComplex));
    memcpy(toReturn.Pointer()+First.Length(),Second.Pointer(),Second.Length()*sizeof(DTDoubleComplex));
    
    return toReturn;
}

DTMutableDoubleComplexArray ConvertToDoubleComplex(const DTDoubleArray &A)
{
    DTMutableDoubleComplexArray toReturn(A.m(),A.n(),A.o());
    int i=0;
    int len = A.Length();
    const double *D = A.Pointer();
    DTDoubleComplex *retD = toReturn.Pointer();
    for (i=0;i<len;i++) {
        retD[i].real() = D[i];
        retD[i].imag() = 0.0;
    }
    return toReturn;
}

DTMutableDoubleArray RealPart(const DTDoubleComplexArray &A)
{
    DTMutableDoubleArray toReturn(A.m(),A.n(),A.o());
    const DTDoubleComplex *D = A.Pointer();
    int i,len = toReturn.Length();
    double *retD = toReturn.Pointer();
    for (i=0;i<len;i++)
        retD[i] = D[i].real();
    
    return toReturn;
}

DTMutableDoubleArray ImaginaryPart(const DTDoubleComplexArray &A)
{
    DTMutableDoubleArray toReturn(A.m(),A.n(),A.o());
    const DTDoubleComplex *D = A.Pointer();
    int i,len = toReturn.Length();
    double *retD = toReturn.Pointer();
    for (i=0;i<len;i++)
        retD[i] = D[i].imag();
    
    return toReturn;
}

