// Part of DTSource. Copyright 2008-2012. David Adalsteinsson.  BSD License
// see http://www.visualdatatools.com/DTSource/license.html for more information.

#ifndef DTDoubleComplexArray_Header
#define DTDoubleComplexArray_Header

#include <iostream>
#include "DTLock.h"

// By default, range check is turned on.
#ifndef DTRangeCheck
#define DTRangeCheck 1
#endif

using namespace std;

#include <complex>

typedef complex<double> DTDoubleComplex;

class DTDoubleComplexArrayStorage {
public:
    DTDoubleComplexArrayStorage(long int mv,long int nv,long int ov);
    ~DTDoubleComplexArrayStorage();

    DTLock accessLock;
    long int m,n,o,mn,length;
    int referenceCount;
    int mutableReferences;
    DTDoubleComplex *Data;
    
private:
    DTDoubleComplexArrayStorage(const DTDoubleComplexArrayStorage &);
    DTDoubleComplexArrayStorage &operator=(const DTDoubleComplexArrayStorage &);
};

class DTMutableDoubleComplexArray;
class DTIndex;
class DTDoubleComplexArrayRegion;
class DTDoubleArray;
class DTMutableDoubleArray;

class DTDoubleComplexArray {

public:
    DTDoubleComplexArray() : Storage(new DTDoubleComplexArrayStorage(0,0,0)), accessLock(), invalidEntry(0.0) {}
    virtual ~DTDoubleComplexArray();
    DTDoubleComplexArray(const DTDoubleComplexArray &A);
    DTDoubleComplexArray &operator=(const DTDoubleComplexArray &A);

protected:
    // If you get a notice that this is protected, change DTDoubleComplexArray to DTMutableDoubleComplexArray
    explicit DTDoubleComplexArray(long int mv,long int nv=1,long int ov=1) : Storage(new DTDoubleComplexArrayStorage(mv,nv,ov)), accessLock(), invalidEntry(0.0) {}

public:
    DTMutableDoubleComplexArray Copy() const;
    
    // Size information.
    long int m() const;
    long int n() const;
    long int o() const;
    long int Length() const;
    bool IsEmpty() const;
    bool NotEmpty() const;
    
    // Low level access
    int ReferenceCount() const;
    int MutableReferences() const; // How many mutable arrays have access to the pointer.
    const DTDoubleComplex *Pointer() const {return Storage->Data;}
    
    // Allow A(i) and A(i,j), but check each access.
#if DTRangeCheck
    DTDoubleComplex operator()(long int i) const
        {if (i<0 || i>=Storage->length)
            {PrintErrorMessage(i); return invalidEntry;}
         return Storage->Data[i];}
    DTDoubleComplex operator()(long int i,long int j) const
        {if (i<0 || i>=Storage->m || j<0 || j>=Storage->n)
            {PrintErrorMessage(i,j); return invalidEntry;}
         return Storage->Data[i+j*Storage->m];}
    DTDoubleComplex operator()(long int i,long int j,long int k) const
        {if (i<0 || i>=Storage->m || j<0 || j>=Storage->n || k<0 || k>=Storage->o)
            {PrintErrorMessage(i,j,k); return invalidEntry;}
            return Storage->Data[i+j*Storage->m+k*Storage->mn];}
#else
    // No range check.  Slightly faster, but not as safe.
    // For fastest access, extract the underlying pointer and dereference it directly.
    DTDoubleComplex operator()(long int i) const {return Storage->Data[i];}
    DTDoubleComplex operator()(long int i,long int j) const {return Storage->Data[i+j*Storage->m];}
    DTDoubleComplex operator()(long int i,long int j,long int k) const {return Storage->Data[i+j*Storage->m+k*Storage->mn];}
#endif

    void pinfo(void) const;

    long int Find(DTDoubleComplex) const; // Returns -1 if not found.
    
    // Support for subregions
    const DTDoubleComplexArrayRegion operator()(DTIndex) const;
    const DTDoubleComplexArrayRegion operator()(DTIndex,DTIndex) const;
    const DTDoubleComplexArrayRegion operator()(DTIndex,DTIndex,DTIndex) const;
    
protected:
    DTDoubleComplexArrayStorage *Storage;
    DTLock accessLock;
    DTDoubleComplex invalidEntry;
    
    // Error messages for index access.
    void PrintErrorMessage(long int i) const;
    void PrintErrorMessage(long int i,long int j) const;
    void PrintErrorMessage(long int i,long int j,long int k) const;
};

class DTMutableDoubleComplexArray : public DTDoubleComplexArray
{
public:
    DTMutableDoubleComplexArray() : DTDoubleComplexArray() {Storage->mutableReferences = 1;}
    ~DTMutableDoubleComplexArray();
    explicit DTMutableDoubleComplexArray(long int mv,long int nv=1,long int ov=1) : DTDoubleComplexArray(mv,nv,ov) {Storage->mutableReferences = 1;}
    DTMutableDoubleComplexArray(const DTMutableDoubleComplexArray &A);

    DTMutableDoubleComplexArray &operator=(const DTMutableDoubleComplexArray &A);

    // Assignment
    DTMutableDoubleComplexArray &operator=(DTDoubleComplex a);

    // Raw access
    DTDoubleComplex *Pointer() {return Storage->Data;}
    const DTDoubleComplex *Pointer() const {return Storage->Data;}

    // High level access
#if DTRangeCheck
    DTDoubleComplex operator()(long int i) const
        {if (i<0 || i>=Storage->length)
            {PrintErrorMessage(i); return invalidEntry;}
            return Storage->Data[i];}
    DTDoubleComplex &operator()(long int i)
        {if (i<0 || i>=Storage->length)
            {PrintErrorMessage(i); return invalidEntry;}
        return Storage->Data[i];}
    DTDoubleComplex operator()(long int i,long int j) const
        {if (i<0 || i>=Storage->m || j<0 || j>=Storage->n)
            {PrintErrorMessage(i,j); return invalidEntry;}
            return Storage->Data[i+j*Storage->m];}
    DTDoubleComplex &operator()(long int i,long int j)
        {if (i<0 || i>=Storage->m || j<0 || j>=Storage->n)
            {PrintErrorMessage(i,j); return invalidEntry;}
        return Storage->Data[i+j*Storage->m];}
    DTDoubleComplex operator()(long int i,long int j,long int k) const
        {if (i<0 || i>=Storage->m || j<0 || j>=Storage->n || k<0 || k>=Storage->o)
            {PrintErrorMessage(i,j,k); return invalidEntry;}
            return Storage->Data[i+j*Storage->m+k*Storage->mn];}
    DTDoubleComplex &operator()(long int i,long int j,long int k)
        {if (i<0 || i>=Storage->m || j<0 || j>=Storage->n || k<0 || k>=Storage->o)
            {PrintErrorMessage(i,j,k); return invalidEntry;}
        return Storage->Data[i+j*Storage->m+k*Storage->mn];}
#else
    DTDoubleComplex operator()(long int i) const {return Storage->Data[i];}
    DTDoubleComplex &operator()(long int i) {return Storage->Data[i];}
    DTDoubleComplex operator()(long int i,long int j) const {return Storage->Data[i+j*Storage->m];}
    DTDoubleComplex &operator()(long int i,long int j) {return Storage->Data[i+j*Storage->m];}
    DTDoubleComplex operator()(long int i,long int j,long int k) const {return Storage->Data[i+j*Storage->m+k*Storage->mn];}
    DTDoubleComplex &operator()(long int i,long int j,long int k) {return Storage->Data[i+j*Storage->m+k*Storage->mn];}
#endif

    // Support for subregions
    const DTDoubleComplexArrayRegion operator()(DTIndex) const;
    const DTDoubleComplexArrayRegion operator()(DTIndex,DTIndex) const;
    const DTDoubleComplexArrayRegion operator()(DTIndex,DTIndex,DTIndex) const;

    DTDoubleComplexArrayRegion operator()(DTIndex);
    DTDoubleComplexArrayRegion operator()(DTIndex,DTIndex);
    DTDoubleComplexArrayRegion operator()(DTIndex,DTIndex,DTIndex);
};

bool operator==(const DTDoubleComplexArray &A,const DTDoubleComplexArray &B);
bool operator!=(const DTDoubleComplexArray &A,const DTDoubleComplexArray &B);

// Misc
extern DTMutableDoubleComplexArray Transpose(const DTDoubleComplexArray &A);
extern DTMutableDoubleComplexArray Reshape(const DTDoubleComplexArray &A,long int m,long int n=1,long int o=1);
extern DTMutableDoubleComplexArray FlipJ(const DTDoubleComplexArray &A);

extern DTMutableDoubleComplexArray CombineColumns(const DTDoubleComplexArray &First,const DTDoubleComplexArray &Second);

// Changing the size of an array
extern DTMutableDoubleComplexArray TruncateSize(const DTDoubleComplexArray &A,long int length);
extern DTMutableDoubleComplexArray IncreaseSize(const DTDoubleComplexArray &A,long int addLength);

// Casting back and forth
extern DTMutableDoubleComplexArray ConvertToDoubleComplex(const DTDoubleArray &);
extern DTMutableDoubleArray RealPart(const DTDoubleComplexArray &);
extern DTMutableDoubleArray ImaginaryPart(const DTDoubleComplexArray &);

#endif
