// Part of DTSource. Copyright 2004-2012. David Adalsteinsson.  BSD License
// see http://www.visualdatatools.com/DTSource/license.html for more information.

#include "DTIndex.h"
#include "DTError.h"

#include "DTIntArrayOperators.h"

const int DTLastIndex = -12563; // Magic number.  Should not use negative indices

// The static variables.
int DTIndex::Last = DTLastIndex;
DTIndex DTIndex::All = DTIndex(0,DTLastIndex);

DTIndex::DTIndex()
: knownLength(-1), start(0), stride(0), end(0), arraySpecified(false), specified()
{
}

DTIndex::DTIndex(int number)
: knownLength(-1), start(0), stride(0), end(0), arraySpecified(false), specified()
{
    // Needs to be either >=0 or <=Last
    if (number<0 && number>DTLastIndex) {
        DTErrorMessage("Subscript error","The subscript was negative.");
        return;
    }

    start = end = number;
    stride = 1;
}

DTIndex::DTIndex(int st,int en)
: knownLength(-1), start(0), stride(0), end(0), arraySpecified(false), specified()
{
    if (st>DTLastIndex && st<0) {
        DTErrorMessage("DTIndex(start,end)","Invalid start (negative)");
        return;
    }
    stride = 1;
    start = st;
    end = en;
}

DTIndex::DTIndex(int start_in,int stride_in,int end_in)
{
    knownLength = -1;
    start = stride = end = 0;
    arraySpecified = false;

    if (start_in>DTLastIndex && start_in<0) {
        DTErrorMessage("DTIndex(start,stride,end)","Invalid start (negative)");
        return;
    }
    if (end_in>DTLastIndex && end_in<0) {
        DTErrorMessage("DTIndex(start,stride,end)","Invalid end (negative)");
        return;
    }
    
    start = start_in;
    stride = stride_in;
    end = end_in;
}

DTIndex::DTIndex(const DTIntArray &spec)
{
    knownLength = -1;
    start = stride = end = 0;
    arraySpecified = true;

    specified = spec;
}

bool DTIndex::SetLength(int howLong)
{
    // Returns false if invalid.
    if (knownLength>=0) {
        DTErrorMessage("Index.SetLength(Len)","Already set the length");
        return false;
    }
    if (howLong<0) {
        DTErrorMessage("Index.SetLength(Len)","Length has to be >= 0");
        return false;
    }

    if (arraySpecified) {
        // Check to make sure that the min and max are within range.
        int i;
        int len = specified.Length();
        for (i=0;i<len;i++) {
            if (specified(i)<0 || specified(i)>=howLong)
                break;
        }
        if (i<len) {
            DTErrorMessage("DTIndex(Array)","Array element out of bounds");
            return false;
        }
    }
    else {
        if (end<=DTLastIndex) {
            // passed in last-k when the index was created.
            end = howLong-1-(DTLastIndex-end);
        }
        if (start<=DTLastIndex) {
            // passed in last-k when the index was created.
            start = howLong-1-(DTLastIndex-start);
        }
        if (end>=howLong || end<0 || start>=howLong || start<0) {
            // start or end outside the current range.
            DTErrorMessage("DTIndex(info)","Index out of range");
            return false;
        }
        if (stride>0 && start>end) {
            // An empty matrix, but this is ok.
            stride = 0;
        }
        else if (stride<0 && start>end) {
            stride = 0;
        }
    }

    knownLength = howLong;

    return true;
}

int DTIndex::Length(void) const
{
    if (arraySpecified)
        return specified.Length(); 
    else
        return (end-start)/stride+1;
}

bool DTIndex::IsEmpty(void) const
{
    if (arraySpecified)
        return specified.IsEmpty();
    return (stride==0);
}

DTIntArray DTIndex::IndicesAsArray(void) const
{
    // Will create it if necessary.
    if (arraySpecified)
        return specified;

    if (stride==0) {
        return DTIntArray();
    }

    int howMany;
    howMany = (end-start)/stride+1;
    int endAt = start + stride*howMany;
    int pos,i = 0;
    DTMutableIntArray toReturn(howMany);
    for (pos=start;pos!=endAt;pos+=stride) {
        toReturn(i++) = pos;
    }
    
    return toReturn;
}

int DTIndex::Start(void) const
{
    return start;
}

int DTIndex::Stride(void) const
{
    return stride;
}

int DTIndex::End(void) const
{
    return end;
}

DTIndex operator-(const DTIndex &A,int b)
{
    if (A.ArraySpecified()) {
        // shift each index
        return DTIndex(A.SpecifiedArray()-b);
    }
    else {
        // start,end,stride
        return DTIndex(A.Start()-b,A.Stride(),A.End()-b);
    }
}

DTIndex operator+(const DTIndex &A,int b)
{
    if (A.ArraySpecified()) {
        // shift each index
        return DTIndex(A.SpecifiedArray()+b);
    }
    else {
        // start,end,stride
        return DTIndex(A.Start()+b,A.Stride(),A.End()+b);
    }
}

