// Part of DTSource. Copyright 2004-2012. David Adalsteinsson.  BSD License
// see http://www.visualdatatools.com/DTSource/license.html for more information.

#include "DTUCharArray.h"
#include "DTError.h"
#include "DTArrayTemplates.h"

#include <string.h>

DTUCharArrayStorage::DTUCharArrayStorage(long int mv,long int nv,long int ov)
{
    // Check if it's called correctly.
    m = mv>0 ? mv : 0;
    n = nv>0 ? nv : 0;
    o = ov>0 ? ov : 0;
    length = m*n*o;
    if (length==0) m = n = o = 0;
    referenceCount = 1;
    mn = m*n;

    Data = length==0 ? NULL : new unsigned char[length];
}

DTUCharArrayStorage::~DTUCharArrayStorage()
{
    delete [] Data;
}

DTUCharArray &DTUCharArray::operator=(const DTUCharArray &A)
{
    // Allow A = A
    if (Storage==A.Storage) return *this;
    
    Storage->accessLock.Lock();
    A.Storage->accessLock.Lock();
    Storage->referenceCount--;
    int refCnt = Storage->referenceCount;
    Storage->accessLock.Unlock();
    if (refCnt==0) delete Storage;
    Storage = A.Storage;
    Storage->referenceCount++;
    Storage->accessLock.Unlock();
    
    return *this;
}

DTMutableUCharArray DTUCharArray::Copy() const
{
    DTMutableUCharArray CopyInto(m(),n(),o());
    // Check that the allocation worked.
    if (CopyInto.Length()!=Length()) return CopyInto; // Failed.  Already printed an error message.
    memcpy(CopyInto.Pointer(),Pointer(),Length()*sizeof(unsigned char));
    return CopyInto;
}

unsigned char DTUCharArray::e(int i) const
{
    if (i<0 || i>=Storage->length) {
        cerr << "Out of bounds" << endl;
        return invalidEntry;
    }
    else
        return Storage->Data[i];
}

unsigned char DTUCharArray::e(int i,int j) const
{
    if (i<0 || i>=Storage->m || j<0 || j>=Storage->n) {
        cerr << "Out of bounds" << endl;
        return invalidEntry;
    }
    else
        return Storage->Data[i+j*Storage->m];
}

unsigned char DTUCharArray::e(int i,int j,int k) const
{
    if (i<0 || i>=Storage->m || j<0 || j>=Storage->n || k<0 || k>=Storage->o) {
        cerr << "Out of bounds" << endl;
        return invalidEntry;
    }
    else
        return Storage->Data[i+j*Storage->m+k*Storage->mn];
}

void DTUCharArray::pinfo(void) const
{
    if (o()==0)
        cerr << "Empty" << endl;
    else if (o()==1) {
        if (n()==1)
            cerr << m() << " entries" << endl;
        else
            cerr << m() << " x " << n() << " array" << endl;
    }
    else
        cerr << m() << " x " << n() << " x " << o() << " array" << endl;
    cerr << flush;
}

void DTUCharArray::pi(int i) const
{
    if (i<0 || i>=m()) {
        cerr << "Out of bounds." << endl;
    }
    else {
        long int howMany = n();
        long int j;
        for (j=0;j<howMany-1;j++) cerr << (int)operator()(i,j) << ", ";
        if (howMany>0) cerr << (int)operator()(i,howMany-1);
        cerr << endl;
    }
}

void DTUCharArray::pj(int j) const
{
    if (j<0 || j>=n()) {
        cerr << "Out of bounds." << endl;
    }
    else {
        long int howMany = m();
        long int i;
        for (i=0;i<howMany-1;i++) cerr << (int)operator()(i,j) << ", ";
        if (howMany>0) cerr << (int)operator()(howMany-1,j);
        cerr << endl;
    }
}

void DTUCharArray::pall(void) const
{
    long int mv = m();
    long int nv = n();
    long int i,j;
    if (mv==0) {
        cerr << "Empty" << endl;
    }
    else {
        for (j=0;j<nv;j++) {
            for (i=0;i<mv-1;i++) cerr << (int)operator()(i,j) << ", ";
            cerr << (int)operator()(mv-1,j);
            cerr << endl;
        }
    }
}

DTMutableUCharArray TruncateSize(const DTUCharArray &A,long int length)
{
    // New length needs to fit as a MxNxO array
    // where MNO = length and
    // if o>1, length = m*n*k
    // if o=1 and n>1 length = m*k
    // if o=1 and n=1 everything is ok.

    if (length==0) return DTMutableUCharArray();
    if (A.IsEmpty()) {
        DTErrorMessage("TruncateSize(Array,Length)","Array is empty.");
        return DTMutableUCharArray();
    }

    long int newM,newN,newO;
    if (A.o()>1) {
        if (length%(A.m()*A.n())!=0) {
            DTErrorMessage("TruncateSize(Array,Length)","Invalid new dimension");
            return DTMutableUCharArray();
        }
        newM = A.m();
        newN = A.n();
        newO = length/(A.m()*A.n());
    }
    else if (A.n()>1) {
        if (length%(A.m())!=0) {
            DTErrorMessage("TruncateSize(Array,Length)","Invalid new dimension");
            return DTMutableUCharArray();
        }
        newM = A.m();
        newN = length/A.m();
        newO = 1;
    }
    else {
        newM = length;
        newN = 1;
        newO = 1;
    }

    DTMutableUCharArray toReturn(newM,newN,newO);
    memcpy(toReturn.Pointer(),A.Pointer(),length*sizeof(unsigned char));
    return toReturn;
}

DTMutableUCharArray IncreaseSize(const DTUCharArray &A,long int addLength)
{
    if (addLength<0) {
        DTErrorMessage("IncreaseSize(Array,Length)","Length needs to be >0.");
        return DTMutableUCharArray();
    }

    long int newM,newN,newO;
    if (A.o()>1) {
        if (addLength%(A.m()*A.n())!=0) {
            DTErrorMessage("IncreaseSize(Array,Length)","Length needs to be a multiple of m*n");
            return DTMutableUCharArray();
        }
        newM = A.m();
        newN = A.n();
        newO = A.o() + addLength/(A.m()*A.n());
    }
    else if (A.n()>1) {
        if (addLength%(A.m())!=0) {
            DTErrorMessage("IncreaseSize(Array,Length)","Length needs to be a multiple of m");
            return DTMutableUCharArray();
        }
        newM = A.m();
        newN = A.n() + addLength/A.m();
        newO = 1;
    }
    else {
        newM = A.m() + addLength;
        newN = 1;
        newO = 1;
    }

    DTMutableUCharArray toReturn(newM,newN,newO);
    memcpy(toReturn.Pointer(),A.Pointer(),A.Length()*sizeof(unsigned char));
    return toReturn;
}

void DTUCharArray::PrintErrorMessage(long int i) const
{
    DTErrorOutOfRange("DTUCharArray",i,Storage->length);
}

void DTUCharArray::PrintErrorMessage(long int i,long int j) const
{
    DTErrorOutOfRange("DTUCharArray",i,j,Storage->m,Storage->n);
}

void DTUCharArray::PrintErrorMessage(long int i,long int j,long int k) const
{
    DTErrorOutOfRange("DTUCharArray",i,j,k,Storage->m,Storage->n,Storage->o);
}

DTMutableUCharArray &DTMutableUCharArray::operator=(unsigned char a)
{
    const long int howManyNumbers = Length();
    long int i;
    unsigned char *Data = Pointer();
    for (i=0;i<howManyNumbers;i++)
        Data[i] = a;
    
    return *this;
}

bool operator==(const DTUCharArray &A,const DTUCharArray &B)
{
    return DTOperatorArrayEqualsArray<DTUCharArray,unsigned char>(A,B);
}

bool operator!=(const DTUCharArray &A,const DTUCharArray &B)
{
    return !(A==B);
}

DTMutableUCharArray Transpose(const DTUCharArray &A)
{
    return DTTransposeArray<DTUCharArray,DTMutableUCharArray,unsigned char>(A);
}

DTMutableUCharArray FlipJ(const DTUCharArray &A)
{
    return DTArrayFlipJ<DTUCharArray,DTMutableUCharArray,unsigned char>(A);
}

void Swap(DTMutableUCharArray &A,DTMutableUCharArray &B)
{
	DTMutableUCharArray C = A;
	A = B;
	B = C;
}

void Swap(DTUCharArray &A,DTUCharArray &B)
{
	DTUCharArray C = A;
	A = B;
	B = C;
}

void CopyValues(DTMutableUCharArray &into,const DTUCharArray &from)
{
	if (into.m()!=from.m() || into.n()!=from.n() || into.o()!=from.o()) {
		DTErrorMessage("CopyValues(MutableUCharArray,UCharArray)","Incompatible sizes");
	}
	else if (into.NotEmpty()) {
		memcpy(into.Pointer(),from.Pointer(),into.Length());
	}
}

