// Part of DTSource. Copyright 2004-2012. David Adalsteinsson.  BSD License
// see http://www.visualdatatools.com/DTSource/license.html for more information.

#ifndef DTDataContainer_Header
#define DTDataContainer_Header

#include "DTDataStorage.h"

#include "DTCharArray.h"
#include "DTDoubleArray.h"
#include "DTFloatArray.h"
#include "DTIntArray.h"
#include "DTShortIntArray.h"
#include "DTUShortIntArray.h"
#include "DTCharArray.h"
#include "DTUCharArray.h"
#include "DTLocked.h"

#include <string>
#include <map>

/*
 This class will handle socket communication with DataTank.  It is still under development, and
 might change frequently.

 To see how it is used, create a project from any of the the "External Program" modules.  You need
 to set the communication method to "Data Streams".
 
 This class is derived from the DTDataStorage base class, just like DTDataFile and DTMatlabDataFile.
 The main difference is that all of the entries are kept in memory.
 
 One usage of this class (separate from DataTank) is to use it as a dynamic data repository where you can
 pass between bigger blocks of your code.  Example:
 
 DTDataContainer inputEntries;
 
 inputEntries = computeInputProperties();
 
 DTDataContainer computed = foo(inputEntries);

 where entries are read and written by using Read and Write, which is overloaded for all of the
 variable types.
 
 */

struct DTDataContainerEntry {
    DTDataContainerEntry() : type(0), sEntry(), dArray(), fArray(), iArray(), sArray(), usArray(), cArray(), ucArray() {}
        
    int type;

    string sEntry;
    DTDoubleArray dArray;
    DTFloatArray fArray;
    DTIntArray iArray;
    DTShortIntArray sArray;
    DTUShortIntArray usArray;
    DTCharArray cArray;
    DTUCharArray ucArray;
    
    int NotFound(void) const {return type==0;}
    int m(void) const;
    int n(void) const;
    int o(void) const;
    
    bool operator!=(const DTDataContainerEntry &) const;
};

class DTDataContainerStorage
{
public:
    DTDataContainerStorage();
    ~DTDataContainerStorage();
    
    int sockfd;
    int communicationSocket;
    bool createdSocket;
    string socketName;
    string responseToDT;
    string requiredAnswer;
    DTMutableCharArray buffer;
    short int portNumber;
    
    bool waitingForClient;
    bool cancelAfterAccept;
    
    bool isReadOnly;
    map<string,DTDataContainerEntry> content;
    DTLock accessLock;
    
    DTLocked<bool> currentlyReading;
    
private:
    DTDataContainerStorage(const DTDataContainerStorage &);
    DTDataContainerStorage &operator=(const DTDataContainerStorage &);
};

class DTDataContainer : public DTDataStorage {
public:
    DTDataContainer();
    
    map<string,DTDataContainerEntry>::const_iterator FirstEntry(void) {return content->content.begin();}
    map<string,DTDataContainerEntry>::const_iterator LastEntry(void) {return content->content.end();}
    
    DTLock &AccessLock(void) {return content->accessLock;}
    bool ConnectToServerSocket(const string &);
    bool ReadFromSocket();
    void RemoveAllEntries();
    bool WriteIntoSocket();
    bool WaitingForClient(void) const {return content->waitingForClient;}
    int NumberOfEntries() {return (content->content.size());}
    
    bool RegisterSocket(const string &);
    bool RegisterSocket(unsigned short int port,const string &responseToDT,const string &requiredAnswer);

    bool WaitForClient(void);
    void TerminateCommunication(void);
    void ForceQuit(void);
    bool ClientConnected(void) const {return (content->communicationSocket>0);}
    void ClientFailed(void);
    bool CurrentlyReading(void) const {return content->currentlyReading;}
    
    int SocketNumber(void) const;
    size_t ReadFromStream(void *,size_t howMuh);
    
    void OverwriteContentWith(const DTDataContainer &);
    
    // Member functions that have to be overwritten.
    DTMutablePointer<DTDataStorage> AsPointer();

    void printInfo(void) const;

    DTList<string> AllVariableNames(void) const;
    bool Contains(const string &name) const;
    bool IsReadOnly(void) const;

    // Saving data.
    void Save(int v,const string &name);
    void Save(double v,const string &name);
    void Save(const DTDoubleArray &A,const string &name);
    void Save(const DTFloatArray &A,const string &name);
    void Save(const DTIntArray &A,const string &name);
    void Save(const DTCharArray &A,const string &name);
    void Save(const DTUCharArray &A,const string &name);
    void Save(const DTShortIntArray &A,const string &name);
    void Save(const DTUShortIntArray &A,const string &name);
    void Save(const string &theString,const string &name);

    bool SavedAsDouble(const string &name) const;
    bool SavedAsString(const string &name) const;

    void Flush(void) const;

    // Reading data.
    DTDoubleArray ReadDoubleArray(const string &name) const;
    DTFloatArray ReadFloatArray(const string &name) const;
    DTIntArray ReadIntArray(const string &name) const;
    DTCharArray ReadCharArray(const string &name) const;
    DTUCharArray ReadUCharArray(const string &name) const;
    DTShortIntArray ReadShortIntArray(const string &name) const;
    DTUShortIntArray ReadUShortIntArray(const string &name) const;
    double ReadNumber(const string &name) const;
    string ReadString(const string &name) const;

    void AddEntry(DTDataContainerEntry,const string &);
    DTDataContainerEntry FindVariable(const string &name) const;

private:
    DTDataContainerEntry FindVariableInsideLock(const string &name) const;
    double ReadNumberInsideLock(const string &name) const;
    string ReadStringInsideLock(const string &name) const;
    DTDoubleArray ReadDoubleArrayInsideLock(const string &name) const;

    DTMutablePointer<DTDataContainerStorage> content;
};


#endif
