// Part of DTSource. Copyright 2004-2012. David Adalsteinsson.  BSD License
// see http://www.visualdatatools.com/DTSource/license.html for more information.

#include "DTMatlabDataFile.h"

#include "DTDoubleArray.h"
#include "DTIntArray.h"
#include "DTUCharArray.h"
#include "DTCharArray.h"
#include "DTFloatArray.h"
#include "DTShortIntArray.h"
#include "DTUShortIntArray.h"
#include "DTUtilities.h"
#include "DTArrayConversion.h"

#include "DTEndianSwap.h"

#include "DTError.h"
#include <string.h>
#include <algorithm>

// The format is specified in the External Interface Guide that comes with Matlab 4.0

// This class does very few checks.
// It won't check if the file opened successfully.
// It won't check to see that you specified a valid file name.

struct FMatrix {
    FMatrix() : type(0), mrows(0), ncols(0), imagf(0), namelen(0) {}
    int type;
    int mrows;
    int ncols;
    int imagf;
    int namelen;
};

struct DTMatlabDataEntry {
    DTMatlabDataEntry() : Fmtrx(), location(-1) {}
    FMatrix Fmtrx;
    long int location; // -1 if not found.
};

class DTMatlabDataFileContent
{
public:
    DTMatlabDataFileContent(DTFile file);
    
    void ReadInContent();
    
    bool needToSwapEndian;
    int referenceCount;
    map<string,DTMatlabDataEntry> content;
    
    DTFile file;
private:
    DTMatlabDataFileContent(const DTMatlabDataFileContent &);
    DTMatlabDataFileContent &operator=(const DTMatlabDataFileContent &);
};

DTMatlabDataFileContent::DTMatlabDataFileContent(DTFile f)
: needToSwapEndian(false), referenceCount(1), file(f)
{
    // Read in the content of the file, so subsequent access is much faster.
    ReadInContent();
}

void DTMatlabDataFileContent::ReadInContent(void)
{
    FMatrix TheMatrix;
    DTMatlabDataEntry singleEntry;

    if (!file.IsOpen())
        return;

    FILE *theFile = file.GetFILE();

    fseek(theFile,0,SEEK_SET);
    size_t howMuchWasRead;
    char tempString[255];
    int IsOK = 1;
    long StartsAt;
    int EntrySize = 0;
    int theSizes[10] = {8,4,4,2,2,1,-1,-1,-1,-1};
    
    int ByteOrderNumber = (DTFile::RunningOnBigEndianMachine() ? 1000 : 0);

    while (IsOK) {
        StartsAt = ftell(theFile); // Where we are currently.

        howMuchWasRead = fread(&TheMatrix,sizeof(FMatrix),1,theFile);
        
        if (howMuchWasRead==0) {
            IsOK = 0;
            break; // The file ended.
        }

        if (StartsAt==0) {
            // Check to see if the endian-ness is different.
            if (TheMatrix.type<ByteOrderNumber || TheMatrix.type>ByteOrderNumber+52) {
                // See if it helps to swap the endian-ness
                DTSwap4Bytes((unsigned char *)&TheMatrix,sizeof(FMatrix));
                needToSwapEndian = true;
            }
        }
        else if (needToSwapEndian) {
            DTSwap4Bytes((unsigned char *)&TheMatrix,sizeof(FMatrix));
        }

        if (howMuchWasRead<1 || TheMatrix.namelen<=0 || TheMatrix.type<0 || TheMatrix.type>1052 || TheMatrix.namelen>255) {
            DTErrorMessage("Reading In File Content",
                           "Invalid file format, make sure you used the -v4 in matlab.");
                if (StartsAt) needToSwapEndian = false;
                IsOK = 0; break;
        }
        
        TheMatrix.type = TheMatrix.type%1000;

        // The variable name
        howMuchWasRead = fread(tempString,1,TheMatrix.namelen,theFile);

        // Figure out the entry size to see how much to skip.
        EntrySize = theSizes[(TheMatrix.type%100)/10]*TheMatrix.mrows*TheMatrix.ncols;
        if (EntrySize<0) {
            DTErrorMessage("dataFile.ReadDoubleArray(name)",
                           "Unexpected variable type in the file.");
            IsOK = 0;
            break;
        }

        if (IsOK) {
            // Add this entry to the content list.
            singleEntry.location = StartsAt;
            singleEntry.Fmtrx = TheMatrix;
            content[string(tempString)] = singleEntry;
        }

        // Else get ready for the next entry.
        fseek(theFile,StartsAt+EntrySize+20+TheMatrix.namelen,SEEK_SET);
    }
}

DTMatlabDataFile::DTMatlabDataFile()
: content(NULL)
{
    DTFile emptyFile;
    content = new DTMatlabDataFileContent(emptyFile);
}

DTMatlabDataFile::DTMatlabDataFile(DTFile file)
: content(NULL)
{
    content = new DTMatlabDataFileContent(file);
}

DTMatlabDataFile::DTMatlabDataFile(const string &name,DTFile::OpenType oType)
: content(NULL)
{
    content = new DTMatlabDataFileContent(DTFile(name,oType));
}

DTMatlabDataFile::DTMatlabDataFile(const DTMatlabDataFile &C)
: content(C.content)
{
    content->referenceCount++;
}

DTMatlabDataFile &DTMatlabDataFile::operator=(const DTMatlabDataFile &C)
{
    if (content==C.content) return *this;
    content->referenceCount--;
    if (content->referenceCount==0)
        delete content;
    content = C.content;
    content->referenceCount++;
    return *this;
}

DTMatlabDataFile::~DTMatlabDataFile()
{
    content->referenceCount--;
    if (content->referenceCount==0)
        delete content;
}

DTMutablePointer<DTDataStorage> DTMatlabDataFile::AsPointer()
{
    return DTMutablePointer<DTDataStorage>(new DTMatlabDataFile(*this));
}

void DTMatlabDataFile::Save(int v,const string &name)
{
    DTMutableDoubleArray temp(1);
    temp(0) = v;
    Save(temp,name);
}

void DTMatlabDataFile::Save(double v,const string &name)
{
    DTMutableDoubleArray temp(1);
    temp(0) = v;
    Save(temp,name);
}

void DTMatlabDataFile::Save(const DTDoubleArray &A,const string &VarName)
{
    FMatrix TheMatrix;

    if (IsReadOnly()) {
        DTErrorMessage("DTMatlabDataFile::Save","File is read only.");
        return;
    }
    
    TheMatrix.type = (DTFile::RunningOnBigEndianMachine() ? 1000 : 0);
    if (A.o()<=1) {
        TheMatrix.mrows = A.m();
        TheMatrix.ncols = A.n();
    }
    else {
        Save(int(A.m()),VarName+"_3D");
        TheMatrix.mrows = A.m()*A.n();
        TheMatrix.ncols = A.o();
    }

    FILE *theFile = GetFile();

    fseek(theFile,0,SEEK_END);
    TheMatrix.imagf = 0;
    TheMatrix.namelen = 1+VarName.length();

    DTMatlabDataEntry entry;
    entry.Fmtrx = TheMatrix;
    entry.location = ftell(theFile);
    content->content[VarName] = entry;

    fwrite(&TheMatrix,sizeof(FMatrix), 1, theFile);
    fwrite(VarName.c_str(),sizeof(char),TheMatrix.namelen,theFile);
    if (A.Length()) fwrite(A.Pointer(),sizeof(double),A.Length(),theFile);
}

void DTMatlabDataFile::Save(const DTFloatArray &A,const string &VarName)
{
    FMatrix TheMatrix;

    if (IsReadOnly()) {
        DTErrorMessage("DTMatlabDataFile::Save","File is read only.");
        return;
    }

    if (A.o()<=1) {
        TheMatrix.mrows = A.m();
        TheMatrix.ncols = A.n();
    }
    else {
        Save(int(A.m()),VarName+"_3D");
        TheMatrix.mrows = A.m()*A.n();
        TheMatrix.ncols = A.o();
    }

    FILE *theFile = GetFile();
    fseek(theFile,0,SEEK_END);

    TheMatrix.type = (DTFile::RunningOnBigEndianMachine() ? 1010 : 10);
    TheMatrix.imagf = 0;
    TheMatrix.namelen = 1+VarName.length();

    DTMatlabDataEntry entry;
    entry.Fmtrx = TheMatrix;
    entry.location = ftell(theFile);
    content->content[VarName] = entry;

    fwrite(&TheMatrix,sizeof(FMatrix), 1, theFile);
    fwrite(VarName.c_str(),sizeof(char),TheMatrix.namelen,theFile);
    if (A.Length()) fwrite(A.Pointer(),sizeof(float),A.Length(),theFile);
}

void DTMatlabDataFile::Save(const DTIntArray &A,const string &VarName)
{
    FMatrix TheMatrix;

    if (IsReadOnly()) {
        DTErrorMessage("DTMatlabDataFile::Save","File is read only.");
        return;
    }

    if (A.o()<=1) {
        TheMatrix.mrows = A.m();
        TheMatrix.ncols = A.n();
    }
    else {
        Save(int(A.m()),VarName+"_3D");
        TheMatrix.mrows = A.m()*A.n();
        TheMatrix.ncols = A.o();
    }

    FILE *theFile = GetFile();
    fseek(theFile,0,SEEK_END);

    TheMatrix.type = (DTFile::RunningOnBigEndianMachine() ? 1020 : 20);
    TheMatrix.imagf = 0;
    TheMatrix.namelen = 1+VarName.length();

    DTMatlabDataEntry entry;
    entry.Fmtrx = TheMatrix;
    entry.location = ftell(theFile);
    content->content[VarName] = entry;

    fwrite(&TheMatrix,sizeof(FMatrix), 1, theFile);
    fwrite(VarName.c_str(),sizeof(char),TheMatrix.namelen,theFile);
    if (A.Length()) fwrite(A.Pointer(),sizeof(int),A.Length(),theFile);
}

void DTMatlabDataFile::Save(const DTUCharArray &A,const string &VarName)
{
    FMatrix TheMatrix;

    if (IsReadOnly()) {
        DTErrorMessage("DTMatlabDataFile::Save","File is read only.");
        return;
    }

    FILE *theFile = GetFile();
    fseek(theFile,0,SEEK_END);

    TheMatrix.type = (DTFile::RunningOnBigEndianMachine() ? 1050 : 50);
    TheMatrix.mrows = A.m();
    TheMatrix.ncols = A.n();
    TheMatrix.imagf = 0;
    TheMatrix.namelen = 1+VarName.length();

    DTMatlabDataEntry entry;
    entry.Fmtrx = TheMatrix;
    entry.location = ftell(theFile);
    content->content[VarName] = entry;

    fwrite(&TheMatrix,sizeof(FMatrix), 1, theFile);
    fwrite(VarName.c_str(),sizeof(char),TheMatrix.namelen,theFile);
    if (A.Length()) fwrite(A.Pointer(),sizeof(char),A.Length(),theFile);
}

void DTMatlabDataFile::Save(const DTCharArray &A,const string &VarName)
{
    FMatrix TheMatrix;

    if (IsReadOnly()) {
        DTErrorMessage("DTMatlabDataFile::Save","File is read only.");
        return;
    }

    FILE *theFile = GetFile();
    fseek(theFile,0,SEEK_END);

    TheMatrix.type = (DTFile::RunningOnBigEndianMachine() ? 1051 : 51);
    TheMatrix.mrows = A.m();
    TheMatrix.ncols = A.n();
    TheMatrix.imagf = 0;
    TheMatrix.namelen = 1+VarName.length();

    DTMatlabDataEntry entry;
    entry.Fmtrx = TheMatrix;
    entry.location = ftell(theFile);
    content->content[VarName] = entry;

    fwrite(&TheMatrix,sizeof(FMatrix), 1, theFile);
    fwrite(VarName.c_str(),sizeof(char),TheMatrix.namelen,theFile);
    if (A.Length()) fwrite(A.Pointer(),sizeof(char),A.Length(),theFile);
}

void DTMatlabDataFile::Save(const DTShortIntArray &A,const string &VarName)
{
    FMatrix TheMatrix;
    
    if (IsReadOnly()) {
        DTErrorMessage("DTMatlabDataFile::Save","File is read only.");
        return;
    }
    
    FILE *theFile = GetFile();
    fseek(theFile,0,SEEK_END);
    
    TheMatrix.type = (DTFile::RunningOnBigEndianMachine() ? 1030 : 30);
    TheMatrix.mrows = A.m();
    TheMatrix.ncols = A.n();
    TheMatrix.imagf = 0;
    TheMatrix.namelen = 1+VarName.length();
    
    DTMatlabDataEntry entry;
    entry.Fmtrx = TheMatrix;
    entry.location = ftell(theFile);
    content->content[VarName] = entry;
    
    fwrite(&TheMatrix,sizeof(FMatrix), 1, theFile);
    fwrite(VarName.c_str(),sizeof(char),TheMatrix.namelen,theFile);
    if (A.Length()) fwrite(A.Pointer(),sizeof(short int),A.Length(),theFile);
}

void DTMatlabDataFile::Save(const DTUShortIntArray &A,const string &VarName)
{
    FMatrix TheMatrix;

    if (IsReadOnly()) {
        DTErrorMessage("DTMatlabDataFile::Save","File is read only.");
        return;
    }

    FILE *theFile = GetFile();
    fseek(theFile,0,SEEK_END);

    TheMatrix.type = (DTFile::RunningOnBigEndianMachine() ? 1040 : 40);
    TheMatrix.mrows = A.m();
    TheMatrix.ncols = A.n();
    TheMatrix.imagf = 0;
    TheMatrix.namelen = 1+VarName.length();

    DTMatlabDataEntry entry;
    entry.Fmtrx = TheMatrix;
    entry.location = ftell(theFile);
    content->content[VarName] = entry;

    fwrite(&TheMatrix,sizeof(FMatrix), 1, theFile);
    fwrite(VarName.c_str(),sizeof(char),TheMatrix.namelen,theFile);
    if (A.Length()) fwrite(A.Pointer(),sizeof(unsigned short int),A.Length(),theFile);
}

void DTMatlabDataFile::Save(const string &theString,const string &VarName)
{
    FMatrix TheMatrix;

    if (IsReadOnly()) {
        DTErrorMessage("DTMatlabDataFile::Save","File is read only.");
        return;
    }
    if (IsOpen()==false) {
        DTErrorMessage("DTMatlabDataFile::Save","File is not open.");
        return;
    }

    FILE *theFile = GetFile();
    fseek(theFile,0,SEEK_END);

    TheMatrix.type = (DTFile::RunningOnBigEndianMachine() ? 1051 : 51); // 8 bit unsigned, and a string.
    TheMatrix.mrows = theString.length();
    TheMatrix.ncols = 1;
    TheMatrix.imagf = 0;
    TheMatrix.namelen = 1+VarName.length();

    DTMatlabDataEntry entry;
    entry.Fmtrx = TheMatrix;
    entry.location = ftell(theFile);
    content->content[VarName] = entry;

    fwrite(&TheMatrix,sizeof(FMatrix), 1, theFile);
    fwrite(VarName.c_str(),sizeof(char),TheMatrix.namelen,theFile);
    fwrite(theString.c_str(),sizeof(char),TheMatrix.mrows,theFile);
}

void DTMatlabDataFile::Flush(void) const
{
    FILE *theFile = GetFile();
    if (theFile) fflush(theFile);
}

DTMatlabDataEntry DTMatlabDataFile::FindVariable(const string &name) const
{
    // should be in the content->content list
    map<string,DTMatlabDataEntry>::const_iterator searchResult = content->content.find(name);
    
    if (searchResult==content->content.end()) {
        return DTMatlabDataEntry();
    }
    else {
        return searchResult->second;
    }
}

DTList<string> DTMatlabDataFile::AllVariableNames(void) const
{
    DTMutableList<string> toReturn(content->content.size());
    
    map<string,DTMatlabDataEntry>::const_iterator mapIterator;
    int pos = 0;
    DTMatlabDataEntry fileEntry;
    
    for (mapIterator=content->content.begin();mapIterator!=content->content.end();++mapIterator) {
        toReturn(pos++) = mapIterator->first;
    }
    
    return toReturn;
}

struct DTMatlabDataFilePosString {
    long int pos;
    string description;

    bool operator<(const DTMatlabDataFilePosString &A) const {return (pos<A.pos);}
};

void DTMatlabDataFile::printInfo(void) const
{
    vector<DTMatlabDataFilePosString> list;
    DTMatlabDataFilePosString entry;
    string desc,stringValue;
    DTMatlabDataEntry fileEntry;

    cerr << "------------------------------------------------------------------------" << endl;
    cerr << "Content of \"" << content->file.Name() << "\" - ";
    int howMany = content->content.size();
    if (howMany==0)
        cerr << "empty" << endl;
    else if (howMany==1)
        cerr << "one entry" << endl;
    else
        cerr << howMany << " entries" << endl;
    cerr << "------------------------------------------------------------------------" << endl;

    string padding = ".................................";

    map<string,DTMatlabDataEntry>::const_iterator mapIterator;
    for (mapIterator=content->content.begin();mapIterator!=content->content.end();++mapIterator) {
        fileEntry = mapIterator->second;
        entry.pos = fileEntry.location;
        desc = mapIterator->first + " ";
        // Pad to make it 30 characters
        if (desc.length()<30)
            desc = desc + string(padding,0,30-desc.length());
        switch (fileEntry.Fmtrx.type%1000) {
            case 0:
                desc += " - double - ";
                break;
            case 10:
                desc += " -  float - ";
                break;
            case 20:
                desc += " -    int - ";
                break;
            case 40:
                desc += " - UShort - ";
                break;
            case 50:
                desc += " -  UChar - ";
                break;
            case 51:
                desc += " - string - ";
                break;
            default:
                desc += " - ?????? - ";
                break;
        }
        // Dimension.
        if (fileEntry.Fmtrx.type%1000==51) {
            stringValue = ReadString(mapIterator->first);
            if (stringValue.length()>25) {
                stringValue = "\""+string(stringValue,0,15) + "...\" - " + DTInt2String(fileEntry.Fmtrx.mrows*fileEntry.Fmtrx.ncols) + " characters";
            }
            desc += "\""+stringValue+"\"";
        }
        else {
            if (fileEntry.Fmtrx.mrows==0)
                desc += "Empty";
            else if (fileEntry.Fmtrx.mrows==1 && fileEntry.Fmtrx.ncols==1)
                desc += DTFloat2StringShort(ReadNumber(mapIterator->first));
            else if (fileEntry.Fmtrx.ncols==1)
                desc += DTInt2String(fileEntry.Fmtrx.mrows) + " numbers";
            else 
                desc += DTInt2String(fileEntry.Fmtrx.mrows) + " x " + DTInt2String(fileEntry.Fmtrx.ncols) + " array";
        }
        entry.description = desc;
        list.push_back(entry);
    }

	std::sort(list.begin(),list.end());

    // Print the content
    int howLong = list.size();
    int pos = 0;
    vector<DTMatlabDataFilePosString>::iterator iter;
    for (iter=list.begin();iter!=list.end();++iter) {
        if (pos<390 || pos>howLong-10)
            cerr << iter->description << endl;
        else if (pos==380 && pos<howLong-20)
            cerr << "Skipping " << howLong-400 << " entries." << endl;
        pos++;
    }
    cerr << flush;
}

bool DTMatlabDataFile::Contains(const string &name) const
{
    DTMatlabDataEntry entry = FindVariable(name);
    return (entry.location>=0);
}

bool DTMatlabDataFile::IsReadOnly() const
{
    return content->file.IsReadOnly();
}

bool DTMatlabDataFile::IsOpen() const
{
    return content->file.IsOpen();
}

bool DTMatlabDataFile::SavedAsDouble(const string &name) const
{
    DTMatlabDataEntry entry = FindVariable(name);
    if (entry.location<0) return false;
    return (entry.Fmtrx.type%1000==0);
}

bool DTMatlabDataFile::SavedAsString(const string &name) const
{
    DTMatlabDataEntry entry = FindVariable(name);
    if (entry.location<0) return false;
    return (entry.Fmtrx.type%1000==51);
}

FILE *DTMatlabDataFile::GetFile(void) const
{
    return content->file.GetFILE();
}

DTDoubleArray DTMatlabDataFile::ReadDoubleArray(const string &name) const
{
    DTMatlabDataEntry entry = FindVariable(name);
    if (entry.location<0) {
        string msg = string("Did not find the variable \"") + name + "\" inside the datafile.";
        DTErrorMessage("dataFile.ReadDoubleArray(name)",msg);
        return DTDoubleArray();
    }

    // See if this is stored as a 3D array - hack since the matlab4 format doesn't
    // directly support it.  An array with size m x n x o
    // is saved as  m*n by o array, and m is saved as the number name_3D
    DTMatlabDataEntry entry3D = FindVariable(name+"_3D");

    FMatrix TheMatrix = entry.Fmtrx;
    long int StartsAt = entry.location;

    int m = TheMatrix.mrows;
    int n = TheMatrix.ncols;
    int o = 1;
    if (entry3D.location>=0) {
        int val = ReadInt(name+"_3D");
        if (val<=0)
            o = 1;
        else {
            m = val;
            n = TheMatrix.mrows/val;
            o = TheMatrix.ncols;
        }
    }

    // Now read the file.
    DTMutableDoubleArray toReturn(m,n,o);
    
    FILE *theFile = GetFile();
    if (TheMatrix.type%1000==0) {
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        if (fread(toReturn.Pointer(),sizeof(double),toReturn.Length(),theFile)!=toReturn.Length()) return DTDoubleArray();
        if (content->needToSwapEndian) SwapEndian(toReturn);
    }
    else if (TheMatrix.type%1000==10) {
        // This is a float arrray.
        DTMutableFloatArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        if (fread(temp.Pointer(),sizeof(float),temp.Length(),theFile)!=temp.Length()) return DTDoubleArray();
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==20) {
        // This is an int arrray.
        DTMutableIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        if (fread(temp.Pointer(),sizeof(int),temp.Length(),theFile)!=temp.Length()) return DTDoubleArray();
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==30) {
        // This is a short arrray.
        DTMutableShortIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        if (fread(temp.Pointer(),sizeof(short),temp.Length(),theFile)!=temp.Length()) return DTDoubleArray();
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==40) {
        // This is an unsigned short arrray.
        DTMutableUShortIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        if (fread(temp.Pointer(),sizeof(short),temp.Length(),theFile)!=temp.Length()) return DTDoubleArray();
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==50) {
        // This is an unsigned char arrray.
        DTMutableCharArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        if (fread(temp.Pointer(),1,temp.Length(),theFile)!=temp.Length()) return DTDoubleArray();
        ConvertArray(temp,toReturn);
    }
    else {
        DTErrorMessage("dataFile.ReadDoubleArray(name)","Illegal type.");
    }

    return toReturn;
}

DTFloatArray DTMatlabDataFile::ReadFloatArray(const string &name) const
{
    DTMatlabDataEntry entry = FindVariable(name);
    if (entry.location<0) {
        string msg = string("Did not find the variable \"") + name + "\" inside the datafile.";
        DTErrorMessage("dataFile.ReadFloatArray(name)",msg);
        return DTFloatArray();
    }

    // See if this is stored as a 3D array - hack since the matlab4 format doesn't
    // directly support it.  An array with size m x n x o
    // is saved as  m*n by o array, and m is saved as the number name_3D
    DTMatlabDataEntry entry3D = FindVariable(name+"_3D");

    FMatrix TheMatrix = entry.Fmtrx;
    long int StartsAt = entry.location;

    int m = TheMatrix.mrows;
    int n = TheMatrix.ncols;
    int o = 1;
    if (entry3D.location>=0) {
        int val = ReadInt(name+"_3D");
        if (val<=0)
            o = 1;
        else {
            m = val;
            n = TheMatrix.mrows/val;
            o = TheMatrix.ncols;
        }
    }

    // Now read the file.
    DTMutableFloatArray toReturn(m,n,o);

    FILE *theFile = GetFile();
    if (TheMatrix.type%1000==0) {
        // This is a double arrray.
        DTMutableDoubleArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(double),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==10) {
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(toReturn.Pointer(),sizeof(float),toReturn.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(toReturn);
    }
    else if (TheMatrix.type%1000==20) {
        // This is an int arrray.
        DTMutableIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(int),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==30) {
        // This is a short arrray.
        DTMutableShortIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(short),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==40) {
        // This is an unsigned short arrray.
        DTMutableUShortIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(short),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==50) {
        // This is an unsigned char arrray.
        DTMutableCharArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),1,temp.Length(),theFile);
        ConvertArray(temp,toReturn);
    }
    else {
        DTErrorMessage("dataFile.ReadFloatArray(name)","Illegal type.");
    }
    
    return toReturn;
}

DTIntArray DTMatlabDataFile::ReadIntArray(const string &name) const
{
    DTMatlabDataEntry entry = FindVariable(name);
    if (entry.location<0) {
        string msg = string("Did not find the variable \"") + name + "\" inside the datafile.";
        DTErrorMessage("dataFile.ReadIntArray(name)",msg);
        return DTIntArray();
    }

    DTMatlabDataEntry entry3D = FindVariable(name+"_3D");
    
    FMatrix TheMatrix = entry.Fmtrx;
    long int StartsAt = entry.location;

    int m = TheMatrix.mrows;
    int n = TheMatrix.ncols;
    int o = 1;
    if (entry3D.location>=0) {
        int val = ReadInt(name+"_3D");
        if (val<=0)
            o = 1;
        else {
            m = val;
            n = TheMatrix.mrows/val;
            o = TheMatrix.ncols;
        }
    }

    // Now read the file.
    DTMutableIntArray toReturn(m,n,o);
    FILE *theFile = GetFile();

    if (TheMatrix.type%1000==0) {
        // This is a double arrray.
        DTMutableDoubleArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(double),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==10) {
        // This is a float arrray.
        DTMutableFloatArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(float),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==20) {
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(toReturn.Pointer(),sizeof(int),toReturn.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(toReturn);
    }
    else if (TheMatrix.type%1000==30) {
        // This is a short arrray.
        DTMutableShortIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(short),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==40) {
        // This is an unsigned short arrray.
        DTMutableUShortIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(short),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==50) {
        // This is an unsigned char arrray.
        DTMutableCharArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),1,temp.Length(),theFile);
        ConvertArray(temp,toReturn);
    }
    else {
        DTErrorMessage("dataFile.ReadIntArray(name)","Illegal type.");
    }
    
    return toReturn;
}

DTUCharArray DTMatlabDataFile::ReadUCharArray(const string &name) const
{
    DTMatlabDataEntry entry = FindVariable(name);
    if (entry.location<0) {
        string msg = string("Did not find the variable \"") + name + "\" inside the datafile.";
        DTErrorMessage("dataFile.ReadUCharArray(name)",msg);
        return DTUCharArray();
    }

    DTMatlabDataEntry entry3D = FindVariable(name+"_3D");
    
    FMatrix TheMatrix = entry.Fmtrx;
    long int StartsAt = entry.location;

    int m = TheMatrix.mrows;
    int n = TheMatrix.ncols;
    int o = 1;
    if (entry3D.location>=0) {
        int val = ReadInt(name+"_3D");
        if (val<=0)
            o = 1;
        else {
            m = val;
            n = TheMatrix.mrows/val;
            o = TheMatrix.ncols;
        }
    }

    // Now read the file.
    DTMutableUCharArray toReturn(m,n,o);
    FILE *theFile = GetFile();

    if (TheMatrix.type%1000==0) {
        // This is a double arrray.
        DTMutableDoubleArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(double),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==10) {
        // This is a float arrray.
        DTMutableFloatArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(float),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==20) {
        // This is an int arrray.
        DTMutableIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(int),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==30) {
        // This is a short arrray.
        DTMutableShortIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(short),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==40) {
        // This is an unsigned short arrray.
        DTMutableUShortIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(short),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==50) {
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(toReturn.Pointer(),sizeof(char),toReturn.Length(),theFile);
    }
    else {
        DTErrorMessage("dataFile.ReadUCharArray(name)","Illegal type.");
    }
    
    return toReturn;
}

DTCharArray DTMatlabDataFile::ReadCharArray(const string &name) const
{
    DTMatlabDataEntry entry = FindVariable(name);
    if (entry.location<0) {
        string msg = string("Did not find the variable \"") + name + "\" inside the datafile.";
        DTErrorMessage("dataFile.ReadCharArray(name)",msg);
        return DTCharArray();
    }

    DTMatlabDataEntry entry3D = FindVariable(name+"_3D");

    FMatrix TheMatrix = entry.Fmtrx;
    long int StartsAt = entry.location;

    int m = TheMatrix.mrows;
    int n = TheMatrix.ncols;
    int o = 1;
    if (entry3D.location>=0) {
        int val = ReadInt(name+"_3D");
        if (val<=0)
            o = 1;
        else {
            m = val;
            n = TheMatrix.mrows/val;
            o = TheMatrix.ncols;
        }
    }

    // Now read the file.
    DTMutableCharArray toReturn(m,n,o);
    FILE *theFile = GetFile();

    if (TheMatrix.type%1000==0) {
        // This is a double arrray.
        DTMutableDoubleArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(double),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==10) {
        // This is a float arrray.
        DTMutableFloatArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(float),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==20) {
        // This is an int arrray.
        DTMutableIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(int),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==30) {
        // This is a short arrray.
        DTMutableShortIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(short),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==40) {
        // This is an unsigned short arrray.
        DTMutableUShortIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(short),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==50 || TheMatrix.type%1000==51) {
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(toReturn.Pointer(),sizeof(char),toReturn.Length(),theFile);
    }
    else {
        DTErrorMessage("dataFile.ReadCharArray(name)","Haven't taken care of this case.");
    }

    return toReturn;
}

DTShortIntArray DTMatlabDataFile::ReadShortIntArray(const string &name) const
{
    DTMatlabDataEntry entry = FindVariable(name);
    if (entry.location<0) {
        string msg = string("Did not find the variable \"") + name + "\" inside the datafile.";
        DTErrorMessage("dataFile.ReadShortIntArray(name)",msg);
        return DTShortIntArray();
    }

    DTMatlabDataEntry entry3D = FindVariable(name+"_3D");

    FMatrix TheMatrix = entry.Fmtrx;
    long int StartsAt = entry.location;

    int m = TheMatrix.mrows;
    int n = TheMatrix.ncols;
    int o = 1;
    if (entry3D.location>=0) {
        int val = ReadInt(name+"_3D");
        if (val<=0)
            o = 1;
        else {
            m = val;
            n = TheMatrix.mrows/val;
            o = TheMatrix.ncols;
        }
    }

    // Now read the file.
    DTMutableShortIntArray toReturn(m,n,o);
    FILE *theFile = GetFile();

    if (TheMatrix.type%1000==0) {
        // This is a double arrray.
        DTMutableDoubleArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(double),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==10) {
        // This is a float arrray.
        DTMutableFloatArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(float),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==20) {
        // This is an int arrray.
        DTMutableIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(int),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==30) {
        // This is a short arrray.
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(toReturn.Pointer(),sizeof(short),toReturn.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(toReturn);
    }
    else if (TheMatrix.type%1000==40) {
        // This is an unsigned short arrray.
        DTMutableUShortIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(short),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==50) {
        // This is an unsigned char arrray.
        DTMutableCharArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),1,temp.Length(),theFile);
        ConvertArray(temp,toReturn);
    }
    else {
        DTErrorMessage("dataFile.ReadShortArray(name)","Illegal type.");
    }
    
    return toReturn;
}

DTUShortIntArray DTMatlabDataFile::ReadUShortIntArray(const string &name) const
{
    DTMatlabDataEntry entry = FindVariable(name);
    if (entry.location<0) {
        string msg = string("Did not find the variable \"") + name + "\" inside the datafile.";
        DTErrorMessage("dataFile.ReadUShortIntArray(name)",msg);
        return DTUShortIntArray();
    }
    
    DTMatlabDataEntry entry3D = FindVariable(name+"_3D");
    
    FMatrix TheMatrix = entry.Fmtrx;
    long int StartsAt = entry.location;
    
    int m = TheMatrix.mrows;
    int n = TheMatrix.ncols;
    int o = 1;
    if (entry3D.location>=0) {
        int val = ReadInt(name+"_3D");
        if (val<=0)
            o = 1;
        else {
            m = val;
            n = TheMatrix.mrows/val;
            o = TheMatrix.ncols;
        }
    }
    
    // Now read the file.
    DTMutableUShortIntArray toReturn(m,n,o);
    FILE *theFile = GetFile();
    
    if (TheMatrix.type%1000==0) {
        // This is a double arrray.
        DTMutableDoubleArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(double),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==10) {
        // This is a float arrray.
        DTMutableFloatArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(float),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==20) {
        // This is an int arrray.
        DTMutableIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(int),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==30) {
        // This is a short arrray.
        DTMutableShortIntArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),sizeof(short),temp.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(temp);
        ConvertArray(temp,toReturn);
    }
    else if (TheMatrix.type%1000==40) {
        // Unsigned short array
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(toReturn.Pointer(),sizeof(unsigned short int),toReturn.Length(),theFile);
        if (content->needToSwapEndian) SwapEndian(toReturn);
    }
    else if (TheMatrix.type%1000==50) {
        // This is an unsigned char arrray.
        DTMutableCharArray temp(m,n,o);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(temp.Pointer(),1,temp.Length(),theFile);
        ConvertArray(temp,toReturn);
    }
    else {
        DTErrorMessage("dataFile.ReadUShortArray(name)","Illegal type.");
    }
    
    return toReturn;
}

double DTMatlabDataFile::ReadNumber(const string &name) const
{
    DTDoubleArray theArr = ReadDoubleArray(name);
    if (theArr.IsEmpty() || theArr.Length()!=1)
        return 0.0;

    return theArr(0);
}

int DTMatlabDataFile::ReadInt(const string &name) const
{
    DTIntArray theArr = ReadIntArray(name);
    if (theArr.IsEmpty() || theArr.Length()!=1)
        return 0;

    return theArr(0);
}

string DTMatlabDataFile::ReadString(const string &name) const
{
    DTMatlabDataEntry entry = FindVariable(name);
    if (entry.location<0) {
        string msg = string("Did not find the variable \"") + name + "\" inside the datafile.";
        DTErrorMessage("dataFile.ReadString(name)",msg);
        return string();
    }

    FMatrix TheMatrix = entry.Fmtrx;
    long int StartsAt = entry.location;

    int theLen = TheMatrix.mrows*TheMatrix.ncols;
    if (theLen==0)
        return string();

    string toReturn;
    if (TheMatrix.type%1000==51) {
        FILE *theFile = GetFile();
        DTMutableUCharArray theBytes(theLen+1);
        fseek(theFile,StartsAt+20+TheMatrix.namelen,SEEK_SET);
        fread(theBytes.Pointer(),sizeof(char),theLen,theFile);
        theBytes(theLen) = 0;
        toReturn = string((char *)theBytes.Pointer());
    }
    else {
        string msg = string("The variable \"") + name + "\" is not save as a string.";
        DTErrorMessage("dataFile.ReadString(name)",msg);
        return string();
    }

    return toReturn;
}

