// Part of DTSource. Copyright 2004-2012. David Adalsteinsson.  BSD License
// see http://www.visualdatatools.com/DTSource/license.html for more information.

#ifndef DTMatlabDataFile_Header
#define DTMatlabDataFile_Header

// A class that can take care of saving arrays in a format that matlab can read.
// use "help load" and "help save" in matlab to get more information.
// On a mac (and maybe also on a pc) you can type
//    load
// in the command window, and drag and drop the file into the window to get
// the full name of the file (including path information).

// Usage is
// Create variables by specifying the file name when they are created.
//
// DTMatlabDataFile TheFile("FileNameOnDisk");
//
// If A is an array (DTDoubleArray or DTFloatArray), to save it in the file
// under the name "ArrayName", use
//     TheFile.Save(A,"ArrayName");
// to read it from the file
//     = TheFile.ReadDoubleArray("ArrayName");
// similarly for other variable types.

// If you are reading or writing variables with names test_1, test_2, ...
// name = "test+" + DTInt2String(3);

#include <stdio.h>
#include <fstream>
#include <map>

#include "DTDataStorage.h"
#include "DTFile.h"

struct DTMatlabDataEntry;
class DTMatlabDataFileContent;

class DTMatlabDataFile : public DTDataStorage {
public:
    DTMutablePointer<DTDataStorage> AsPointer();

    DTMatlabDataFile();
    DTMatlabDataFile(DTFile file);
    // DTFile::ReadOnly, DTFile::ExistingReadWrite, DTFile::NewReadWrite
    DTMatlabDataFile(const string &name,DTFile::OpenType=DTFile::ExistingReadWrite);

    // Copying and assignment treat this class as a pointer.
    DTMatlabDataFile(const DTMatlabDataFile &);
    DTMatlabDataFile &operator=(const DTMatlabDataFile &);
    ~DTMatlabDataFile();

    bool Contains(const string &name) const;
    DTList<string> AllVariableNames(void) const;
    bool IsReadOnly(void) const;
    FILE *GetFile(void) const;
    bool IsOpen(void) const;
  
    // Saving data.
    void Save(int,const string &name);
    void Save(double,const string &name);
    void Save(const DTDoubleArray &A,const string &name);
    void Save(const DTFloatArray &A,const string &name);
    void Save(const DTIntArray &A,const string &name);
    void Save(const DTCharArray &A,const string &name);
    void Save(const DTUCharArray &A,const string &name);
    void Save(const DTShortIntArray &A,const string &name);
    void Save(const DTUShortIntArray &A,const string &name);
    void Save(const string &theString,const string &name);

    bool SavedAsDouble(const string &name) const;
    bool SavedAsString(const string &name) const;

    void Flush(void) const;

    // Reading data.
    DTDoubleArray ReadDoubleArray(const string &name) const;
    DTFloatArray ReadFloatArray(const string &name) const;
    DTIntArray ReadIntArray(const string &name) const;
    DTCharArray ReadCharArray(const string &name) const;
    DTUCharArray ReadUCharArray(const string &name) const;
    DTShortIntArray ReadShortIntArray(const string &name) const;
    DTUShortIntArray ReadUShortIntArray(const string &name) const;
    double ReadNumber(const string &name) const;
    int ReadInt(const string &name) const;
    string ReadString(const string &name) const;
    
private:
    void printInfo(void) const;
    DTMatlabDataEntry FindVariable(const string &name) const;

    DTMatlabDataFileContent *content;
};

#endif
