// Part of DTSource. Copyright 2004-2012. David Adalsteinsson.  BSD License
// see http://www.visualdatatools.com/DTSource/license.html for more information.

#include "DTStringList.h"

#include "DTError.h"
#include <string.h>

DTStringList::DTStringList(const DTCharArray &chars,const DTIntArray &offs)
{
    // Check validity.
    if (chars.IsEmpty() && offs.IsEmpty())
        return;
    
    if (chars.n()!=1 || chars.o()!=1 || offs.n()!=1 || offs.o()!=1) {
        DTErrorMessage("DTStringList(characters,offsets)","Invalid array sizes.");
        return;
    }
    
    // offsets need to be between 0 and len-1, and increasing.
    int numberOfChars = chars.Length();
    int howMany = offs.Length();
    int i,off;
    for (i=0;i<howMany;i++) {
        off = offs(i);
        if (off<0 || off>=numberOfChars) break;
    }
    if (i<howMany) {
        DTErrorMessage("DTStringList(characters,offsets)","One of the offsets is out of range.");
        return;
    }
    
    // Increasing
    for (i=1;i<howMany;i++) {
        if (offs(i-1)>offs(i)) break;
    }
    if (i<howMany) {
        DTErrorMessage("DTStringList(characters,offsets)","The offsets need to be increasing.");
        return;
    }
    
    characters = chars;
    offsets = offs;
}

DTStringList::DTStringList(const DTList<string> &entries)
{
    // Figure out the total length
    int totalLength = 0;
    int howManyStrings = entries.Length();
    int i;
    for (i=0;i<howManyStrings;i++) {
        totalLength += entries(i).length()+1;
    }
    
    DTMutableIntArray offs(howManyStrings);
    DTMutableCharArray chars(totalLength);
    
    int posInChars = 0;
    int singleLength;
    for (i=0;i<howManyStrings;i++) {
        offs(i) = posInChars;
        singleLength = entries(i).length()+1;
        memcpy(chars.Pointer()+posInChars,entries(i).c_str(),singleLength);
        posInChars += singleLength;
    }
    
    characters = chars;
    offsets = offs;
}
    
string DTStringList::operator()(int si) const
{
    if (si<0 || si>=offsets.Length()) {
        DTErrorMessage("DTStringList::SingleEntry","Out of bounds.");
        return string();
    }
    
    int theLength;
    if (si<offsets.Length()-1)
        theLength = offsets(si+1)-offsets(si);
    else
        theLength = characters.Length()-offsets(si);
    
    if (theLength<=1)
        return string();
    
    return string(characters.Pointer()+offsets(si),theLength-1);
}

void DTStringList::pinfo(void) const
{
    if (offsets.Length()==0) {
        cerr << "No strings\n";
    }
    else if (NumberOfStrings()<100) {
        int i;
        int howMany = NumberOfStrings();
        for (i=0;i<howMany;i++) {
            cerr << operator()(i) << endl;
        }
    }
    else {
        cerr << offsets.Length() << " strings\n";
    }
    cerr << flush;
}

void Read(const DTDataStorage &input,const string &name,DTStringList &toReturn)
{
    DTIntArray offsets;
    DTCharArray characters;
    
    Read(input,name+"_offs",offsets);
    Read(input,name,characters);
    
    toReturn = DTStringList(characters,offsets);
}

void Write(DTDataStorage &output,const string &name,const DTStringList &theVar)
{
    Write(output,name+"_offs",theVar.Offsets());
    Write(output,name,theVar.Characters());
}

void WriteOne(DTDataStorage &output,const string &name,const DTStringList &toWrite)
{
    Write(output,name,toWrite);
    Write(output,"Seq_"+name,"List of Strings");
    output.Flush();
}

